/* Representation of DHCPv6 lease options for libdhcp */

/*
 * Copyright (C) 2006  Red Hat, Inc. All rights reserved.
 *
 * This copyrighted material is made available to anyone wishing to use,
 * modify, copy, or redistribute it subject to the terms and conditions of
 * the GNU General Public License v.2, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY expressed or implied, including the implied warranties of
 * MERCHANTABILITY or FITNESS FOR A * PARTICULAR PURPOSE.  See the GNU General
 * Public License for more details.  You should have received a copy of the
 * GNU General Public License along with this program; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
 * source code or documentation are not subject to the GNU General Public
 * License and may only be used or replicated with the express permission of
 * Red Hat, Inc.
 *
 * Red Hat Author(s): Jason Vas Dias
 *                    David Cantrell <dcantrell@redhat.com>
 */

#ifndef _DHCP6_LEASE_H
#define _DHCP6_LEASE_H

#include  <sys/types.h>
#include  <netinet/in.h>
#include  <stdint.h>

#include  <sys/queue.h>    /* glibc list macros */

/** @addtogroup DHCPv6_lease 
 *  @{
 */

/** 
 *  DHCPv6 address types.
 */
enum dhcpv6_address_type_e
{
    DHCPv6_ADDRESS,           /**< permanent lease address */
    DHCPv6_TEMPORARY_ADDRESS, /**< temporary address given with delegated prefix */
    DHCPv6_DELEGATED_PREFIX   /**< delegated prefix to configure radvd(8) with */
}   DHCPv6_address_type;

/**
 *  DHCPv6 address type.
 *  There may be multiple DHCPv6 addresses per lease,
 *  each of which may have different prefix length,
 *  and lifetimes.
 */
typedef
struct dhcp6_lease_address_s
{
    struct in6_addr address;        /**< the IPv6 lease address bytes */
    uint8_t         prefix_length;  /**< lease address prefix length */ 
    uint8_t         type;           /**< lease address type */
    uint16_t        status;         /**< lease address status */
    time_t          time_acquired;  /**< time lease address acquired */
    time_t          valid_lifetime; /**< valid lifetime of lease address */
    time_t          prefer_lifetime;/**< prefer lifetime of lease address */
    STAILQ_ENTRY    ( dhcp6_lease_address_s ) link; /**< next lease address */
}   DHCPv6_lease_address;


/**
 * DHCPv6 lease address list type.
 */
typedef
STAILQ_HEAD( dhcp6_lease_address_list_s, dhcp6_lease_address_s ) 
DHCPv6_lease_address_list;

/**
 * DHCPv6 DNS server IPv6 address list element type.
 */
typedef
struct ipv6_address_s
{
    struct in6_addr address;
    STAILQ_ENTRY( ipv6_address_s ) link;
}   IPv6_address;

/**
 * DHCPv6 DNS server IPv6 address list type.
 */
typedef 
STAILQ_HEAD( ipv6_address_list_s, ipv6_address_s )
IPv6_address_list;


/**
 * DHCPv6 lease type
 */
typedef 
struct dhcp6_lease_s
{
    DHCPv6_lease_address_list addresses;  /**< list of lease addresses */
    time_t   renew_time;                  /**< time lease should be renewed */
    time_t   rebind_time;                 /**< time lease must be rebound */
    char     *search_list;                /**< dns domain name search list, if sent */
    IPv6_address_list dns;                /**< domain name servers */ 
    char     *if_name;                    /**< interface name      */
    uint32_t  if_index;                   /**< interface index    */
    uint32_t iaid;  		          /**< DHCPv6 lease identifier */ 
    uint8_t     *client_duid;  	          /**< DHCPv6 client identifier */
    uint8_t     *server_duid;  	          /**< DHCPv6 server identifier */
    uint16_t     client_duid_len;         /**< DHCPv6 client identifier length*/
    uint16_t     server_duid_len;  	  /**< DHCPv6 server identifier length*/
    struct in6_addr server_address;       /**< DHCPv6 server address */
} DHCPv6_lease;

struct dhcp6_optinfo;           /**< defined in dhcpv6's dhcp6.h */

extern DHCPv6_lease *dhcpv6_lease ( struct dhcp6_optinfo * );/**<
  * call this with the dhcp6_optinfo * returned to the
  * libdhcp callback on the DHC6_BOUND state to construct
  * a DHCPv6_lease .
  */

extern void dhcpv6_lease_free( DHCPv6_lease * ); /**<
  * frees all resources associated with the DHCPv6_lease
  */

extern int dhcpv6_pack_lease( DHCPv6_lease* lease, uint8_t* buf, uint32_t len); /**<
  * packs the DHCPv6_lease in a buffer of length len, suitable for IPC 
  * / mmap'ed file storage.
  */

extern DHCPv6_lease *dhcpv6_unpack_lease( uint8_t* buf ); /**<
  * unpacks the DHCPv6_lease from a buffer created by dhcpv6_pack_lease()
  */

/**@}*/

#endif
