/*
 *  API for libdhcp, a minimal interface to the ISC dhcp IPv4 client,
 *  and to the dhcpv6 DHCPv6 client libraries, libdhcp4client, and 
 *  libdhcp6client .
 */

/*
 * Copyright (C) 2006  Red Hat, Inc. All rights reserved.
 *
 * This copyrighted material is made available to anyone wishing to use,
 * modify, copy, or redistribute it subject to the terms and conditions of
 * the GNU General Public License v.2, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY expressed or implied, including the implied warranties of
 * MERCHANTABILITY or FITNESS FOR A * PARTICULAR PURPOSE.  See the GNU General
 * Public License for more details.  You should have received a copy of the
 * GNU General Public License along with this program; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
 * source code or documentation are not subject to the GNU General Public
 * License and may only be used or replicated with the express permission of
 * Red Hat, Inc.
 *
 * Red Hat Author(s): Jason Vas Dias
 *                    David Cantrell <dcantrell@redhat.com>
 */

#ifndef LIBDHCP_H
#define LIBDHCP_H

#include <sys/types.h>
#include <sys/syslog.h>
#include <stdint.h>
#include <stdarg.h>

/**
 * this must be defined during the DHCP client compile to
 * select libdhcp specific modicifications:
 */
#define LIBDHCP 1

/**
 * The DHCP_state enum represents the states of the DHCP clients,
 * that each client will call the user's call back for on a state change.
 */
typedef enum dhcp_state_e
{

    /* DHCPv4 client states - third callback arg will be a 'struct client_state *'    */
    DHC4_NBI,         		/**< failed: no broadcast interfaces found              */
    DHC4_PREINIT, 		/**< configuration started - bring the interface "UP"   */
    DHC4_BOUND, 		/**< lease obtained                                     */
    DHC4_RENEW, 		/**< lease renewed                                      */
    DHC4_REBOOT,	        /**< have valid lease, but now obtained a different one */
    DHC4_REBIND, 		/**< new, different lease                               */
    DHC4_STOP,  		/**< remove old lease                                   */
    DHC4_MEDIUM, 		/**< media selection begun                              */
    DHC4_TIMEOUT, 		/**< timed out contacting DHCP server                   */
    DHC4_FAIL, 			/**< all attempts to contact server timed out, sleeping */
    DHC4_EXPIRE, 		/**< lease has expired, renewing                        */
    DHC4_RELEASE, 		/**< releasing lease                                    */
    /* This state raised by both clients: */
    DHC_TIMEDOUT,               /**< libdhcp_control timeout has been exceeded          */
    /* DHCPv6 client states:    */
    DHC6_BOUND,                 /**< new lease obtained             - arg is optinfo *  */
    DHC6_REBIND,                /**< existing expired lease rebound - arg is optinfo *  */
    DHC6_RELEASE                /**< existing lease expired         - arg is dhcp6_iaidaddr*/
} DHCP_State;

struct libdhcp_control_s;

typedef 
int ( *LIBDHCP_Callback ) ( struct libdhcp_control_s *control, enum dhcp_state_e state, void* arg );/**<
* The DHCP client callback type.
* The DHCP clients will call the users' callback on important state change events,
* which is of this function type, and is passed in through the client control structure.
* @arg control:    the DHCP client control handle.
* @arg state:      the client DHCP_State.
* @arg arg:        user pointer from the control structure. 
*/

#define LOG_FATAL 8
typedef
int ( *LIBDHCP_Error_Handler )
    ( struct libdhcp_control_s *ctl,
      int priority,
      const char *fmt,
      va_list ap
    ); /**<
 * The libdhcp error handler / logger callback type.
 * The DHCP clients, and parts of libdhcp, that need to report / log an error condition will
 * call a function of this type, passed in through the control structure.
 * No DHCP client or libdhcp code will do any logging unless a non-zero pointer to a function
 * of this type is recorded in the ctl->eh member.
 * @arg ctl:        the DHCP client control handle.
 * @arg priority:   the log level / priority of the log message, ala syslog(3) priority.
 * The priority corresponds to the syslog priorities LOG_ERR, LOG_INFO, and LOG_DEBUG.
 * A special LOG_FATAL priority (8), not used by syslog, is used to indicate a condition
 * that would cause the client to exit when not run under libdhcp. LOG_FATAL messages 
 * are always logged, and the handler sets ctl->finished = 1 on receipt of a LOG_FATAL
 * message - @see LIBDHCP_Control::finished.
 */

/** LIBDHCP_Control
 * 
 *  The libdhcp DHCP client control structure.
 *  Each client is called with a pointer to this structure as the first argument
 *  to its main dhcp*_client function.
 *  This structure specifies what the client is allowed to do and how it will
 *  communicate state changes to the calling code.
 */
typedef 
struct libdhcp_control_s
{
/*{ this structure MUST be the same as in libdhcp_control.h for DHCP clients */
    LIBDHCP_Callback    callback;    	/**< the DHCP clients' main loop calls this on state changes */
    uint16_t            capability;     /**< LIBDHCP_Capability bits to enable                       */
    uint8_t             finished;       /**< set to one to make clients exit their main loop         */
    uint8_t             decline;        /**< set to one to decline the lease (DHCPv4 only)           */
    time_t              timeout;        /**< (timeout+now) == time after which clients MUST return   */
    time_t              now;            /**< clients set this to time(0) on entering main loop       */
    void               *arg;            /**< user data pointer                                       */
    LIBDHCP_Error_Handler eh;           /**< The logger / error hander callback                      */
/*} end of fields known to DHCP clients */
    uint8_t             log_level;      /**< maximum log level (LOG_FATAL excluded)                  */  
} LIBDHCP_Control;

/** 
 * LIBDHCP_Capability 
 *
 * enum that defines the capabilities DHCP clients are allowed to exercise.
 *
 * The DHCP_CONFIGURE* capabilities are supported by DHCPv6 in process, 
 * while the DHCPv4 client would fork and exec the dhclient-script to implement them if these
 * bits are set - otherwise, if no bits are set, the callback is called and the script is 
 * not run. 
 * @warning If using the dhcp4_nic DHCPv4 network interface configurator, DO NOT set any of
 * the DHCP_CONFIGURE_* bits to one.
 */
typedef enum libdhcp_capability_e
{
    DHCP_USE_LEASE_DATABASE   = 1,  	/* use / do not use persistent lease database files   */
    DHCP_USE_PID_FILE         = 2,  	/**< use / do not use pid file                        */
    DHCP_CONFIGURE_INTERFACES = 4,  	/**< configure interfaces UP/DOWN as required         */
    DHCP_CONFIGURE_ADDRESSES  = 8,  	/**< configure interface addresses as required        */
    DHCP_CONFIGURE_ROUTES     =16,  	/**< configure routes as required                     */
    DHCP_CONFIGURE_RESOLVER   =32,  	/**< configure resolv.conf as required                */
    DHCP_CONFIGURE_RADVD      =64,  	/**< (DHCPv6 only) configure radvd.conf & restart radvd as required */
} LIBDHCP_Capability;

/**
 * @var  LIBDHCP_Capability::DHCP_USE_LEASE_DATABASE
 * @brief use / do not use persistent lease database files.
 * If this bit is set to one, clients will maintain a persistent lease database,
 * as they would under normal DHCP client operation.
 * When restarted, they will first check for existing unexpired leases in the lease database
 * and will request the latest unexpired lease from the server; if no server can be contacted,
 * they will enter the REBIND state with the latest unexpired lease.
 * The IPv4 client (DHCPv4) will maintain its lease database in /var/lib/dhclient ; 
 * the IPv6 client (DHCPv6) will maintain its lease database in /var/lib/dhcpv6 .
 * @warning 
 * Use of DHCP_USE_LEASE_DATABASE with DHCPv6 when invoking dhcpv6_client multiple
 * times in the same process will lead to memory leaks in the flex scanner used for parsing
 * the DHCPv6 lease database .
 * This problem is being worked on, but is expected most users of the dhcp6client library
 * will be invoking the client only once / not using DHCP_USE_LEASE_DATABASE .
 * DHCPv4 has no memory leaks.
 */

/**
 * libdhcp_control_new - creates a new DHCP client control LIBDHCP_Control structure.
 * The arguments correspond to the named LIBDHCP_Control structure members .
 * @see LIBDHCP_control 
 */
extern LIBDHCP_Control*
libdhcp_control_new
(   LIBDHCP_Callback   	   callback,      /**< the user's function to be called on state changes*/
    LIBDHCP_Capability     dhc_cap,       /**< DHCP client capabilities to enable - @see LIBDHCP_Capability*/
    time_t                 timeout,       /**< time period in seconds after which client must return */
    void                  *arg,           /**< argument to pass to user's callback as third arg */
    LIBDHCP_Error_Handler  error_handler, /**< the logger this client is to use */
    uint8_t                log_level      /**< the maximum log level to be logged by this client */
);  

extern void libdhcp_control_free(LIBDHCP_Control *);

/**
 * The function prototype of the DHCP client main() function or primary entry point, 
 * (renamed dhcpv4_client for the DHCPv4 client, and dhcpv6_client for the DHCPv6 client).
 * Users must include the appropriate dhcpv{4,6}client.h header to obtain the declarations
 * of these clients' entry points.
 */
typedef int (*DHCP_Client) ( LIBDHCP_Control*, int argc, char **argv, char **envp );

extern int
libdhcp_call_client( LIBDHCP_Control *,  DHCP_Client, ...); /**<
 * calls the dhcpv{4,6}_client function with the arguments in the variable argument list.
 * @arg ctl : the client control structure - @see LIBDHCP_Control
 * @arg client: the dhcp client to run - @see DHCP_Client
 * @arg ... : The argv[] argument list to pass to the client.
 * The last argument in the argument list MUST be 0 .
 * The dhcpv4 client has many useful arguments to pass, 
 * eg. the \b -H \b <host-name> , \b -V \b <vendor-class-identifier>,
 * and \b -I \b <dhcp-client-identifier> arguments - see man dhclient(8).  
 */

/**
 * function to convert a ::DHCP_State to a string.
 * @arg buf : if non-zero, will be used to store the string
 * @warning If buf is 0, the string returned is to a static buffer,
 * and the call cannot be used more than once in an argument list.
 */
char *libdhcp_state_string( DHCP_State, char *buf );

/**
 * default error handler that logs to stderr.
 * @see LIBDHCP_Error_Handler
 */
extern 
int libdhcp_stderr_logger
(   struct libdhcp_control_s *ctl,
    int priority,
    const char *fmt,
    va_list ap
);

/**
 * default error handler that logs to syslog.
 * @see LIBDHCP_Error_Handler
 */
extern 
int libdhcp_syslogger
(   struct libdhcp_control_s *ctl,
    int priority,
    const char *fmt,
    va_list ap
);

/** @} */

#endif

