/*
 * Network Interface Configurator (NIC) for libdhcp.
 *
 * Provides facilities for configuring network interfaces, addresses,
 * and routes, with an interface to the libnl netlink library.
 */

/*
 * Copyright (C) 2006  Red Hat, Inc. All rights reserved.
 *
 * This copyrighted material is made available to anyone wishing to use,
 * modify, copy, or redistribute it subject to the terms and conditions of
 * the GNU General Public License v.2, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY expressed or implied, including the implied warranties of
 * MERCHANTABILITY or FITNESS FOR A * PARTICULAR PURPOSE.  See the GNU General
 * Public License for more details.  You should have received a copy of the
 * GNU General Public License along with this program; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
 * source code or documentation are not subject to the GNU General Public
 * License and may only be used or replicated with the express permission of
 * Red Hat, Inc.
 *
 * Red Hat Author(s): Jason Vas Dias
 *                    David Cantrell <dcantrell@redhat.com>
 */

#ifndef LIBDHCP_NIC_H
#define LIBDHCP_NIC_H

#include <sys/types.h>
#include <sys/queue.h>
#include <netinet/in.h>
#include <net/if.h>
#include <stdint.h>
#include <stdarg.h>

#include "ip_addr.h"

/* Network Interface Configuration (NIC) Link opaque type */
struct nic_s;
typedef struct nic_s *NIC_t;

/* Interface IP Address opaque type */
struct  nic_ip_address_s;
typedef struct nic_ip_address_s *IPaddr_t;

typedef struct nic_ip_address_list_node_s {
    IPaddr_t addr;
    STAILQ_ENTRY(nic_ip_address_list_node_s) link;
} IPaddr_list_node_t;

typedef STAILQ_HEAD(nic_ip_address_list_s,nic_ip_address_list_node_s) IPaddr_list_t;

/* Interface Route opaque type */
struct nic_route_s;
typedef struct nic_route_s *IProute_t;

typedef struct nic_ip_route_node_s {
    IProute_t route;
    STAILQ_ENTRY(nic_ip_route_node_s) link;
} IProute_list_node_t;

typedef STAILQ_HEAD(nic_ip_route_list_s,nic_ip_route_node_s) IProute_list_t;

/*
 * Opaque NIC Library Handle - users must first obtain one to create 
 * any NIC object. Also records error handler and log level:
 */
typedef struct nlh_s *NLH_t;

/*
 * Error level - functions will call error handler
 * with priority set to one of these values.
 */
typedef enum nic_error_level_e {
    NIC_FATAL,      /* cannot continue - unused!   */
    NIC_ERR = 3,    /* operation failed            */
    NIC_WARN = 4,   /* not all OK with operation   */
    NIC_DEBUG = 7   /* debugging                   */
} NIC_Error_Level_t;

/* Result code type */
typedef enum nic_result_e {
    NIC_FAIL = -1,   /* operation failed */
    NIC_OK   =  0,   /* success */
    NIC_SUCCESS = 0  /* success == NIC_OK */
} NIC_Res_t;

/* Error handler / logger function type */
typedef void (*NIC_Error_Handler_t)(NLH_t, NIC_Error_Level_t, char*, ...);

/* First call this to obtain a NIC module "handle" */
extern NLH_t nic_open(NIC_Error_Handler_t);

/* Call this when finished with NIC to free resources */
extern void nic_close(NLH_t *);


/* default error handlers */
extern void nic_set_logger(NLH_t, NIC_Error_Handler_t);
typedef int (*NIC_VA_Error_Handler_t) (void *, int priority, char *fmt,
                                       va_list va);
extern void nic_set_va_logger(NLH_t, NIC_VA_Error_Handler_t handler,
                              void *handler_arg);
extern void nic_sys_logger(NLH_t, NIC_Error_Level_t, char *fmt, ...);
extern void nic_stderr_logger(NLH_t, NIC_Error_Level_t, char *fmt, ...);
extern NIC_Res_t nic_set_loglevel(NLH_t, NIC_Error_Level_t);

/* Interface NIC (Link / IF) functions */
extern NIC_t nic_by_name(NLH_t, char *);
extern NIC_t nic_by_index(NLH_t, int16_t);

typedef void (*NIC_handler_t)(NLH_t nh, NIC_t, void*);

/*
 * Calls handler with nic of each interface currently configured in
 * the kernel and user void* arg arguments.
 * @warning: the callback MUST NOT free(nic)!
 * Also, only ONE thread can do a nic foreach at a time - this is enforced.
 */
extern void nic_foreach(NLH_t nh, NIC_handler_t handler,  void *arg);

extern char *nic_get_name(NIC_t);
extern int16_t nic_get_index(NIC_t);

/* get link layer / ethernet address */
extern ip_addr_t nic_get_link_addr(NIC_t);

/* get link layer broadcast address */
extern ip_addr_t nic_get_link_broadcast(NIC_t);

/*
 * call nic_update(nic) after you've made
 * all your nic_set_*(nic,...) calls to
 * actually update the link with netlink.
 */
extern NIC_Res_t nic_update(NIC_t nic);
extern uint32_t nic_get_flags(NIC_t);
extern void nic_set_flags(NIC_t, uint32_t);
extern uint32_t nic_get_mtu(NIC_t);
extern void nic_set_mtu(NIC_t, uint32_t);
extern char *nic_get_qdisc(NIC_t);
extern void nic_set_qdisc(NIC_t, char *);
extern uint32_t nic_get_txqlen(NIC_t);
extern void nic_set_txqlen(NIC_t, uint32_t);
extern uint32_t nic_get_link(NIC_t);
extern void nic_set_link(NIC_t, uint32_t);
extern uint32_t nic_get_weight(NIC_t);
extern void nic_set_weight(NIC_t, uint32_t);
extern uint32_t nic_get_master(NIC_t);
extern void nic_set_master(NIC_t, uint32_t);
extern uint32_t nic_get_cost(NIC_t);
extern void nic_set_cost(NIC_t, uint32_t);
extern uint32_t nic_get_priority(NIC_t);
extern void nic_set_priority(NIC_t, uint32_t);
extern uint32_t nic_get_protinfo(NIC_t);
extern void nic_set_protinfo(NIC_t, uint32_t);

struct rtnl_link_stats;
extern struct rtnl_link_stats nic_get_stats(NIC_t);

struct rtnl_link_ifmap;
extern struct rtnl_link_ifmap nic_get_ifmap(NIC_t);

/*
 * NIC IP Address functions:
 */
typedef void (*IPaddr_Handler_t) (NLH_t nh, IPaddr_t, void *);
extern void nic_addr_foreach(NLH_t, IPaddr_Handler_t, void *);
extern IPaddr_t nic_addr_ip(NLH_t nh, ip_addr_t *);
extern ip_addr_t nic_ip_addr(IPaddr_t);
extern IPaddr_t nic_addr(NLH_t nh, ip_addr_t);
extern IPaddr_t nic_addr_local(NLH_t nh, ip_addr_t);
extern ip_addr_t nic_addr_get_local(IPaddr_t);
extern void nic_addr_set_local(IPaddr_t, ip_addr_t);
extern uint8_t nic_addr_get_family(IPaddr_t);
extern uint8_t nic_addr_get_prefix(IPaddr_t);
extern void nic_addr_set_prefix(IPaddr_t, uint8_t);
extern ip_addr_t nic_addr_get_broadcast(IPaddr_t);
extern void nic_addr_set_broadcast(IPaddr_t, ip_addr_t);
extern ip_addr_t nic_addr_get_anycast(IPaddr_t);
extern void nic_addr_set_anycast(IPaddr_t, ip_addr_t);
extern ip_addr_t nic_addr_get_multicast(IPaddr_t);
extern void nic_addr_set_multicast(IPaddr_t, ip_addr_t);
extern int8_t nic_addr_get_scope(IPaddr_t);
extern void nic_addr_set_scope(IPaddr_t, int8_t);
extern uint8_t nic_addr_get_flags(IPaddr_t);
extern void nic_addr_set_flags(IPaddr_t, uint8_t);
extern const char *nic_addr_get_label(IPaddr_t);
extern void nic_addr_set_label(IPaddr_t, const char *);

struct ifa_cacheinfo;
extern struct ifa_cacheinfo nic_addr_get_cacheinfo(IPaddr_t);

extern void  nic_addr_set_cacheinfo(IPaddr_t, struct ifa_cacheinfo*);

extern IPaddr_list_t *nic_address_list_new(IPaddr_t, ...);
extern void nic_address_list_free(IPaddr_list_t *);
extern NIC_Res_t nic_add_address(NIC_t, IPaddr_t);
extern NIC_Res_t nic_remove_address(NIC_t, IPaddr_t);
extern NIC_Res_t nic_add_addresses(NIC_t, IPaddr_list_t *);
extern NIC_Res_t nic_remove_addresses(NIC_t, IPaddr_list_t *);
void nic_addr_free(void *);

/*
 * NIC IP Route functions: 
 */
typedef void (*IProute_handler_t) (NLH_t nh, IProute_t, void *);

extern void nic_route_foreach(NLH_t, IProute_handler_t, void *);
extern int32_t nic_route_get_table(IProute_t);
extern void nic_route_set_table(IProute_t, uint8_t table);
extern char *nic_route_get_table_name(int32_t, char *buf, int len);
extern int32_t nic_route_get_table_number(char *);
extern uint8_t nic_route_get_family(IProute_t);
extern uint8_t nic_route_get_scope(IProute_t);
extern void nic_route_set_scope(IProute_t, uint8_t);
extern uint32_t nic_route_get_flags(IProute_t);
extern void nic_route_set_flags(IProute_t, uint32_t);
extern uint8_t nic_route_get_dst_len(IProute_t);
extern void nic_route_set_dst_len(IProute_t, uint8_t);
extern uint8_t nic_route_get_src_len(IProute_t);
extern void nic_route_set_src_len(IProute_t, uint8_t);
extern uint8_t nic_route_get_type(IProute_t);
extern void nic_route_set_type(IProute_t, uint8_t);
extern uint8_t nic_route_get_protocol(IProute_t);
extern void nic_route_set_protocol(IProute_t, uint8_t);
extern uint8_t nic_route_get_tos(IProute_t);
extern void nic_route_set_tos(IProute_t, uint8_t);
extern ip_addr_t nic_route_get_dst(IProute_t);
extern void nic_route_set_dst(IProute_t, ip_addr_t);
extern ip_addr_t nic_route_get_src(IProute_t);
extern void nic_route_set_src(IProute_t, ip_addr_t);
extern ip_addr_t nic_route_get_gateway(IProute_t);
extern void nic_route_set_gateway(IProute_t, ip_addr_t);
extern ip_addr_t nic_route_get_prefsrc(IProute_t);
extern void nic_route_set_prefsrc(IProute_t, ip_addr_t);
extern int16_t nic_route_get_oif(IProute_t);
extern void nic_route_set_oif(IProute_t, uint16_t);

typedef struct nic_if_name_s {
    char name[IFNAMSIZ];
} NIC_if_name_t;

extern NIC_if_name_t nic_route_get_iif(IProute_t);
extern void nic_route_set_iif(IProute_t, char *);
extern uint32_t nic_route_get_priority(IProute_t);
extern void nic_route_set_priority(IProute_t, uint32_t);
extern uint32_t nic_route_get_protoinfo(IProute_t);
extern void nic_route_set_protoinfo(IProute_t, uint32_t);
extern uint32_t nic_route_get_session(IProute_t);
extern void nic_route_set_session(IProute_t, uint32_t);
extern uint32_t nic_route_get_flow(IProute_t);
extern void nic_route_set_flow(IProute_t, uint32_t);
extern uint32_t nic_route_get_mp_algo(IProute_t);
extern void nic_route_mp_algo(IProute_t, uint32_t);
#define NIC_ROUTE_DEFAULT_ADDR ((ip_addr_t *) NULL)
#define NIC_ROUTE_NO_IF (-1)
extern IProute_t nic_route_new(NLH_t, int16_t oif, ip_addr_t *dst,
                               uint8_t dst_len, ip_addr_t *gw, int8_t type,
                               int8_t protocol, int8_t scope,
                               int32_t priority, int32_t table, char* iif,
                               ip_addr_t *src, uint8_t src_len);
extern void nic_route_free(void *);
extern NIC_Res_t nic_add_route(IProute_t);
extern NIC_Res_t nic_update_route(IProute_t);
extern NIC_Res_t nic_remove_route(IProute_t);
extern IProute_list_t *nic_route_list_new(IProute_t, ...);
void nic_route_list_free(IProute_list_t *);
extern NIC_Res_t nic_add_routes(IProute_list_t *);
extern NIC_Res_t nic_remove_routes(IProute_list_t *);
extern NIC_Res_t nic_update_routes(IProute_list_t *);
extern void nic_free_route(IProute_t);
struct rtnl_route *nic_rtnl_route_get(IProute_t);

/* Every other RTA attribute can be set with 
 * the libnl rtnl_route module - include the
 * libnl headers to use, and call rtnl_route_put
 * when finished with the pointer.
 */
void nic_rtnl_route_put(struct rtnl_route *);

/*
 * Global Configuration Operations:
 */
extern NIC_Res_t nic_set_hostname(NLH_t nh, char *hostname);
extern NIC_Res_t nic_configure(NLH_t nh, NIC_t nic, IPaddr_list_t *addresses,
                               IProute_list_t *routes,
                               IPaddr_list_t *dns_servers,
                               char *search_list, char *host_name);
#endif
