/*
 * Network Interface Configurator for BOTH the ISC DHCP IPv4 client library
 * and the DHCPv6 IPv6 client.
 *
 * Copyright (C) 2006  Red Hat, Inc. All rights reserved.
 *
 * This copyrighted material is made available to anyone wishing to use,
 * modify, copy, or redistribute it subject to the terms and conditions of
 * the GNU General Public License v.2, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY expressed or implied, including the implied warranties of
 * MERCHANTABILITY or FITNESS FOR A * PARTICULAR PURPOSE.  See the GNU General
 * Public License for more details.  You should have received a copy of the
 * GNU General Public License along with this program; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
 * source code or documentation are not subject to the GNU General Public
 * License and may only be used or replicated with the express permission of
 * Red Hat, Inc.
 *
 * Red Hat Author(s): Jason Vas Dias
 *                    David Cantrell <dcantrell@redhat.com>
 */

#include <sys/types.h>
#include <unistd.h>
#include <sys/wait.h>
#include <sys/mman.h>
#include <string.h>
#include <malloc.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>

#include "dhcp_nic.h"

static int dhc_log(LIBDHCP_Control *ctl, int priority, char *fmt, ...) {
    if ((ctl==0) || (ctl->eh == 0))
        return 0;

    va_list va;
    va_start(va,fmt);
    int r = ctl->eh(ctl, priority, fmt, va);
    va_end(va);
    return r;
}

static int dhc_log_noctl(LIBDHCP_Error_Handler eh, int log_level, int priority,
                         char *fmt, ...) {        
    LIBDHCP_Control ctl;

    if (eh == 0)
        return 0;

    ctl.log_level = log_level;
    va_list va;
    va_start(va,fmt);
    int r = eh(&ctl, priority, fmt, va);
    va_end(va);
    return r;
}

static int ifup(LIBDHCP_Control *control, NIC_t nic) {
    /* bring interface UP */
    uint32_t iff = nic_get_flags(nic);

    dhc_log(control, LOG_INFO, "DHCP %s - bringing link UP - %x %d",
            nic_get_name(nic), iff, iff & (IFF_UP | IFF_RUNNING));

    if ((iff & (IFF_UP | IFF_RUNNING)) != (IFF_UP | IFF_RUNNING)) {
        nic_set_flags(nic, iff | IFF_UP | IFF_RUNNING);

        if (nic_update(nic) != NIC_OK) {
            dhc_log(control, LOG_ERR, "DHCP %s - bringing link UP failed.",
                    nic_get_name(nic));
            return 1;
        }
    }

    return 0;
}

static DHCP_nic *dhcp_nic_va(NLH_t nh, DHCP_Preference preference, char *ethX,
                             LIBDHCP_Capability dhc_cap, time_t timeout,
                             LIBDHCP_Error_Handler error_handler,
                             uint8_t log_level, va_list dhclient_va) {
    DHCP_nic *nic = calloc(1, sizeof(DHCP_nic));

    if (nic == NULL) {
        dhc_log_noctl(error_handler, log_level, LOG_FATAL,
                      "dhcp_nic: out of memory");
        return NULL;
    }

    nic->preference = preference;
    nic->nh = nh;
    NIC_t eth = nic_by_name(nh, ethX);

    if (eth == 0) {
        dhc_log_noctl(error_handler, log_level, LOG_FATAL,
                      "dhcp_nic: net_get_by_name(%s) failed", ethX);
        free(nic);
        return NULL;
    }

    LIBDHCP_Control *dhc6ctl = NULL;
    DHCPv4_control *dhc4ctl = NULL;

    if ((preference & DHCPv4_DISABLE) == 0) {
        dhc4ctl = dhcpv4_control_va(nh, ethX, dhc_cap, timeout, error_handler,
                                    log_level, dhclient_va);
        if (dhc4ctl == NULL) {
            dhc_log_noctl(error_handler, log_level, LOG_FATAL,
                          "dhcp_nic: dhcpv4_control failed");
            free(nic);
            return NULL;
        }

        nic->dhc4ctl = dhc4ctl;
    }

    if ((preference & DHCPv6_DISABLE) == 0) {
        dhc6ctl = libdhcp_control_new(dhcp6_nic_callback, 0, timeout, 0,
                                      error_handler, log_level);
        if (dhc6ctl == NULL) {
            dhc_log_noctl(error_handler, log_level, LOG_FATAL,
                          "dhcp_nic: libdhcp_control_new failed");
            free(nic);
            if (dhc4ctl)
                dhcpv4_control_free(dhc4ctl);
            return NULL;
        }

        nic->dhc6ctl = dhc6ctl;
    }

    if ((dhc6ctl == NULL) && (dhc4ctl == NULL)) {
        dhc_log_noctl(error_handler, log_level, LOG_FATAL,
                      "dhcp_nic: both DISABLE_DHCPv6 and DISABLE_DHCv4 set.");
        free(nic);
        return NULL;
    }

    if (error_handler) {
        nic_set_va_logger(nh, (NIC_VA_Error_Handler_t) error_handler, dhc6ctl);
        nic_set_loglevel(nh, log_level);
    }

    if (ifup(dhc6ctl, eth) != 0) {
        dhc_log_noctl(error_handler, log_level, LOG_FATAL,
                      "dhcp_nic: ifup(%s) failed", ethX);

        if (dhc4ctl)
            dhcpv4_control_free(dhc4ctl);
        if (dhc6ctl)
            libdhcp_control_free(dhc6ctl);

        free(nic);
        return NULL;
    }

    DHCPv4_nic *nic4 = NULL;
    DHCPv6_nic *nic6 = NULL;

    if ((preference & DHCPv6_DISABLE) == 0) {
        /*
         * Ensure the ipv6 kernel module is loaded -
         * it gets auto-loaded by first process to do this:
         */
        int ipv6_sock = socket(AF_INET6, SOCK_DGRAM, IPPROTO_UDP);
        close(ipv6_sock);
    }

    if (timeout > 0) {
        if (preference & IPv6_PREFERENCE) {
            if ((preference & DHCPv6_DISABLE) == 0)
                nic6 = do_dhcpv6(nic->dhc6ctl, nh, ethX);

            if ((preference & DHCPv4_DISABLE) == 0)
                nic4 = do_dhcpv4(nic->dhc4ctl);
        } else {
            if ((preference & DHCPv4_DISABLE) == 0)
                nic4 = do_dhcpv4(nic->dhc4ctl);

            if ((preference & DHCPv6_DISABLE) == 0)
            nic6 = do_dhcpv6(nic->dhc6ctl, nh, ethX);
        }
    } else {
        /* User has selected timeout==0, so each client will continue
         * trying to obtain a lease indefinitely. This can only work
         * in separate processes / threads - threads don't work because
         * of the select() used by both clients, so we use processses.
         */

        pid_t dhc4_pid = -1, dhc6_pid = -1;

        /* 128k - there may be LOTS of DHCPv4 options */
        uint8_t *sma = mmap(0, 16 * 8192, PROT_READ | PROT_WRITE,
                            MAP_SHARED | MAP_ANONYMOUS, -1, 0);
        if (sma == 0) {
            dhc_log_noctl(error_handler, log_level, LOG_FATAL, "dhcp_nic: timeout==0 and cannot obtain shared memory area - %d %s.", errno, strerror(errno));

            if (dhc4ctl)
                dhcpv4_control_free(dhc4ctl);
            if (dhc6ctl)
                libdhcp_control_free(dhc6ctl);

            free(nic);
            return NULL;
        }

        uint8_t *dhc4_buf = sma, *dhc6_buf = sma + (14 * 8192);

        if ((preference & DHCPv4_DISABLE) == 0) {
            if ((dhc4_pid = fork()) == 0 ) {
                nic4 = do_dhcpv4(nic->dhc4ctl);

                if (nic4 && nic4->lease != 0)
                    dhcpv4_pack_lease(nic4->lease, dhc4_buf, (14 * 8192));

                if (dhc4ctl)
                    dhcpv4_control_free(dhc4ctl);
                if (dhc6ctl)
                    libdhcp_control_free(dhc6ctl);

                free(nic);
                exit(0);
            }

            if (dhc4_pid == -1) {
                dhc_log_noctl(error_handler, log_level, LOG_FATAL, "dhcp_nic: timeout==0 and fork() failed - %d %s.", errno, strerror(errno));

                if (dhc4ctl)
                    dhcpv4_control_free(dhc4ctl);
                if (dhc6ctl)
                    libdhcp_control_free(dhc6ctl);

                free(nic);
                return 0L;
            }
        }

        if ((preference & DHCPv6_DISABLE) == 0) {
            if ((dhc6_pid = fork()) == 0) {
                nic6 = do_dhcpv6(nic->dhc6ctl, nh, ethX);

                if (nic6 && nic6->lease)
                    dhcpv6_pack_lease(nic6->lease, dhc6_buf, (2 * 8192));

                if (dhc4ctl)
                    dhcpv4_control_free(dhc4ctl);
                if (dhc6ctl)
                    libdhcp_control_free(dhc6ctl);

                free(nic);
                exit(0);
            }

            if (dhc6_pid == -1) {
                dhc_log_noctl(error_handler, log_level, LOG_FATAL, "dhcp_nic: timeout==0 and fork() failed - %d %s.", errno, strerror(errno));

                if (dhc4ctl)
                    dhcpv4_control_free(dhc4ctl);
                if (dhc6ctl)
                    libdhcp_control_free(dhc6ctl);

                free(nic);
                return NULL;
            }
        }

        pid_t wait_pid = 0;
        int status, dhc4_status = 0, dhc6_status = 0;

dhc_client_wait:
        dhc_log_noctl(error_handler, log_level, LOG_DEBUG, "DHCP: wait - %d %d", dhc4_status, dhc6_status);

        if ((dhc4_status == 0 || dhc6_status == 0) ||
            (dhc6_pid == -1 && dhc4_pid == -1))
            goto dhc_clients_finished;

        wait_pid = waitpid(-1, &status, 0);

        if (wait_pid == -1)
            goto dhc_client_wait;

        if (wait_pid == dhc4_pid) {
            DHCPv4_lease *lease4 = dhcpv4_unpack_lease(dhc4_buf);
            dhc4_status = dhc4_pid;

            dhc_log_noctl(error_handler, log_level, LOG_DEBUG,
                          "DHCPv4 process finished - %p", lease4);

            if (lease4)
                nic4 = dhcp4_set_lease(dhc4ctl, lease4);

            if (dhc6_status == 0) {
                if (lease4 && (preference & DHCP_ACCEPT_FIRST_LEASE)) {
                    kill(dhc6_pid, SIGKILL);
                    waitpid(dhc6_pid, 0, WNOHANG);
                } else {
                    goto dhc_client_wait;
                }
            }
        } else {
            if (wait_pid == dhc6_pid) {
                DHCPv6_lease *lease6 = dhcpv6_unpack_lease(dhc6_buf);
                dhc6_status = dhc6_pid;

                dhc_log_noctl(error_handler, log_level, LOG_DEBUG, "DHCPv6 process finished - %p", lease6);

                if (lease6)
                    nic6 = dhcp6_nic_from_lease(dhc6ctl, nh, lease6, eth);

                if (dhc4_status == 0) {
                    if (lease6 && (preference & DHCP_ACCEPT_FIRST_LEASE)) {
                        kill(dhc4_pid, SIGKILL);
                        waitpid(dhc4_pid, 0, WNOHANG);
                    } else {
                        goto dhc_client_wait;
                    }
                }
            } else {
                goto dhc_client_wait;
            }
        }

dhc_clients_finished:
        munmap(sma, 16 * 8192);
    }

    if (!dhc4ctl) {
        nic4 = NULL;
        nic->dhc4ctl = NULL;
    }

    if (nic4 && !dhcp4_process_lease(dhc4ctl)) {
        dhc_log_noctl(error_handler, log_level, LOG_ERR, "DHCP: empty DHCPv4 lease.");
        dhcpv4_control_free(dhc4ctl);
        dhc4ctl = nic->dhc4ctl = NULL;
        nic4 = NULL;
    }

    if (nic6 && !dhcp6_process_lease(dhc6ctl, nic6)) {
        dhc_log_noctl(error_handler, log_level, LOG_ERR, "DHCP: empty DHCPv6 lease.");
        dhcpv6_nic_free(nic6);
        nic6 = NULL;
        free(dhc6ctl);
        nic->dhc6ctl = dhc6ctl = NULL;
    }

    if (!nic6 && !nic4) {
#if 0
        if (dhc4ctl)
            dhcpv4_control_free(dhc4ctl);
        if (dhc6ctl)
            libdhcp_control_free(dhc6ctl);
#endif
        free(nic);
        return NULL;
    }

    nic->dhcpv4 = (DHCP_config*) nic4;
    nic->dhcpv6 = (DHCP_config*) nic6;
    return nic;
}

DHCP_nic *dhcp_nic(NLH_t nh, DHCP_Preference preference, char *ethX,
                   LIBDHCP_Capability dhc_cap, time_t timeout,
                   LIBDHCP_Error_Handler error_handler,
                   uint8_t log_level, ...) {
    va_list dhclient_va;
    va_start(dhclient_va, log_level);
    DHCP_nic *nic = dhcp_nic_va(nh, preference, ethX, dhc_cap, timeout,
                                error_handler, log_level, dhclient_va);
    va_end(dhclient_va);
    return nic;
}

void dhcp_nic_free(DHCP_nic *nic) {
    if (nic->dhc4ctl) {
        dhcpv4_control_free(nic->dhc4ctl);
        nic->dhc4ctl = NULL;
    }

    if (nic->dhc6ctl) {
        free(nic->dhc6ctl);
        nic->dhc6ctl = NULL;
    }

    if (nic->dhcpv6) {
        dhcpv6_nic_free((DHCPv6_nic*) nic->dhcpv6);
        nic->dhcpv6 = NULL;
    }

    if (nic->nh) {
        nic_close(&nic->nh);
    }

    free(nic);
    return;
}

NIC_Res_t do_dhcp(DHCP_Preference preference, char *eth_if_name,
                  LIBDHCP_Capability dhc_cap, time_t timeout,
                  LIBDHCP_Error_Handler error_handler,
                  uint8_t log_level, ...) {
    va_list dhclient_va;
    va_start(dhclient_va, log_level);

    NLH_t nh = nic_open(0);
    if (nh == 0) {
        dhc_log_noctl(error_handler, log_level, LOG_FATAL,
                      "do_dhcp: nic_open() failed");
        nic_close(&nh);
        return 0;
    }

    DHCP_nic *nic = dhcp_nic_va(nh, preference, eth_if_name, dhc_cap, timeout,
                                error_handler, log_level, dhclient_va);
    va_end(dhclient_va);

    if (nic == 0) {
        dhc_log_noctl(error_handler, log_level, LOG_FATAL,
                      "do_dhcp: no leases obtained.");
        nic_close(&nh);
        return NIC_FAIL;
    }

    NIC_Res_t r = dhcp_nic_configure(nic);
    dhcp_nic_free(nic);
    nic_close(&nh);

    return r;
}

NIC_Res_t dhcp_nic_configure(DHCP_nic *nic) {
    if (nic == 0)
        return NIC_FAIL;

    LIBDHCP_Control *ctl = nic->dhc6ctl;
    NIC_Res_t res = NIC_FAIL;
    char order[3] = { 0, 0, 0 };
    char res_order[3] = { 0, 0, 0 };
    IPaddr_list_t *dns = NULL;
    char *search_list = NULL;
    int i;
    IPaddr_list_node_t *n = NULL, *nn = NULL;
    IPaddr_list_t *addrlist = NULL;
    IProute_list_t *routelist = NULL;
    char *host_name = NULL;

    if (nic->dhcpv6) {
        if ((!(nic->preference & DHCPv6_DISABLE_RESOLVER)) &&
            !STAILQ_EMPTY(&nic->dhcpv6->dns_list)) {
            res_order[0] = 6;
        }

        order[0] = 6;
    }

    if (nic->dhcpv4) {
        if ((!(nic->preference & DHCPv6_DISABLE_RESOLVER)) &&
            !STAILQ_EMPTY(&nic->dhcpv4->dns_list)) {
            res_order[1] = 4;
        }

        order[1] = 4;
    }

    if (!(nic->preference & IPv6_PREFERENCE)) {
        char a = order[0];
        order[0] = order[1];
        order[1] = a;

        a = res_order[0];
        res_order[0] = res_order[1];
        res_order[1] = res_order[0];
    }

    for (i = 0; i < 2; i++) {
        if (!order[i])
            order[i] = order[i+1];
        if (!res_order[i])
            res_order[i] = res_order[i+1];
    }

    dns = calloc(1, sizeof(IPaddr_list_t));
    STAILQ_INIT(dns);

    for (i = 0; res_order[i]; i++) {
        if (search_list != NULL) {
            free(search_list);
            search_list = NULL;
        }

        switch (res_order[i]) {
            case 6:
                if (!nic->dhcpv6) 
                    continue;
                STAILQ_FOREACH(n, &(nic->dhcpv6->dns_list), link) {
                    nn = calloc(1, sizeof(IPaddr_list_node_t));
                    nn->addr = n->addr;
                    STAILQ_INSERT_TAIL(dns, nn, link);
                }

                if (nic->dhcpv6->search_list)
                    search_list = strdup(nic->dhcpv6->search_list);

                break;
            case 4:
                if (!nic->dhcpv4)
                    continue;
                STAILQ_FOREACH(n, &(nic->dhcpv4->dns_list), link) {
                    nn = calloc(1, sizeof(IPaddr_list_node_t));
                    nn->addr = n->addr;
                    STAILQ_INSERT_TAIL(dns, nn, link);
                }

                if (nic->dhcpv4->search_list)
                    search_list = strdup(nic->dhcpv4->search_list);

                break;
        }
    }

    for (i = 0; order[i]; i++) {
        switch (order[i]) {
            case 6:
                if (!nic->dhcpv6) 
                    continue;

                if (nic->preference & DHCPv6_DISABLE_ADDRESSES)
                    addrlist = NULL;
                else
                    addrlist = &nic->dhcpv6->address_list;

                res = nic_configure(nic->nh, nic->dhcpv6->nic, addrlist, 0,
                                    dns, search_list, 0);

                if (res == NIC_OK)
                    dhc_log(ctl, LOG_INFO, "DHCPv6 interface configuration succeeded.");
                else
                    dhc_log(ctl, LOG_ERR, "DHCPv6 interface configuration failed.");
                break;
            case 4:
                if (!nic->dhcpv4)
                    continue;

                /* DHCPv4 */
                if (nic->dhc4ctl && dhcpv4_mtu_option(nic->dhc4ctl)) {
                    nic_set_mtu(nic->dhcpv4->nic,
                        dhcpv4_mtu_option(nic->dhc4ctl));

                    if (nic_update(nic->dhcpv4->nic) != NIC_OK)
                        dhc_log(ctl, LOG_ERR, "DHCPv4 MTU set failed."); 
                }

                if (nic->preference & DHCPv4_DISABLE_ADDRESSES)
                    addrlist = NULL;
                else
                    addrlist = &nic->dhcpv4->address_list;

                if (nic->preference & DHCPv4_DISABLE_ROUTES)
                    routelist = NULL;
                else
                    routelist = &nic->dhcpv4->route_list;

                if (nic->preference & DHCPv4_DISABLE_HOSTNAME_SET)
                    host_name = NULL;
                else
                    host_name = nic->dhcpv4->host_name;

                res = nic_configure(nic->nh, nic->dhcpv4->nic, addrlist,
                                    routelist, dns, search_list, host_name);

                if (res == NIC_OK)
                    dhc_log(ctl, LOG_INFO, "DHCPv4 interface configuration succeeded.");
                else
                    dhc_log(ctl, LOG_ERR, "DHCPv4 interface configuration failed.");
                break;
        }
    }

    if (!STAILQ_EMPTY(dns)) {
        nn = STAILQ_FIRST( dns );
        while ((n = nn)) {
            nn = STAILQ_NEXT(n, link);
            free(n);
        }
    }

    if (dns)
        free(dns);

    if (search_list)
        free(search_list);

    return res;
}
