/* Generic IP address representation and conversion utilities. */

/*
 * Copyright (C) 2006  Red Hat, Inc. All rights reserved.
 *
 * This copyrighted material is made available to anyone wishing to use,
 * modify, copy, or redistribute it subject to the terms and conditions of
 * the GNU General Public License v.2, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY expressed or implied, including the implied warranties of
 * MERCHANTABILITY or FITNESS FOR A * PARTICULAR PURPOSE.  See the GNU General
 * Public License for more details.  You should have received a copy of the
 * GNU General Public License along with this program; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
 * source code or documentation are not subject to the GNU General Public
 * License and may only be used or replicated with the express permission of
 * Red Hat, Inc.
 *
 * Red Hat Author(s): Jason Vas Dias
 *                    David Cantrell <dcantrell@redhat.com>
 */

#include <arpa/inet.h>
#include <net/if_arp.h>
#include <string.h>
#include <stdio.h>

#include "ip_addr.h"

ip_addr_t ip_addr_v4(register uint32_t i) {
    ip_addr_t ip = { AF_INET };
    memset(&(ip.sa_data[0]),'\0',sizeof(ip.sa_data));
    ((struct in_addr*)IP_ADDR_IN(&ip))->s_addr = htonl(i);
    return ip;
}

uint32_t ip_v4_addr(register ip_addr_t *ip) {
    if (ip->sa_family != AF_INET)
        return 0;
    return ntohl(((struct in_addr*) IP_ADDR_IN(ip))->s_addr);
}

ip_addr_t ip_addr_in(register struct in_addr *in) {
    ip_addr_t ip = { AF_INET };
    memset(&(ip.sa_data[0]),'\0',sizeof(ip.sa_data));
    *((struct in_addr*) IP_ADDR_IN(&ip)) = *in;
    return ip;
}

struct in_addr ip_in_addr(register ip_addr_t *ip) {
    return *((struct in_addr*) IP_ADDR_IN(ip));
}

ip_addr_t ip_addr_sin(register struct sockaddr_in *sin) {
    ip_addr_t ip;
    *((struct sockaddr_in*) &ip) = *sin;
    return ip;
}

struct sockaddr_in ip_sin_addr(register ip_addr_t *ip) {
    return *((struct sockaddr_in*) ip);
}

ip_addr_t ip_addr_in6(register struct in6_addr *in6) {
    ip_addr_t ip = { AF_INET6 };
    memset(&(ip.sa_data[0]),'\0',sizeof(ip.sa_data));
    *((struct in6_addr*) IP_ADDR_IN6(&ip)) = *in6;
    return ip;
}

struct in6_addr ip_in6_addr(register ip_addr_t *ip) {
    return *((struct in6_addr*) IP_ADDR_IN6(ip));
}

ip_addr_t ip_addr_sin6(register struct sockaddr_in6 *sin6) {
    ip_addr_t ip;
    *((struct sockaddr_in6*) &ip) = *sin6;
    return ip;
}

struct sockaddr_in6 ip_sin6_addr(register ip_addr_t *ip) {
    return *((struct sockaddr_in6*) ip);
}

ip_addr_t ip_addr_in6bytes(register in6_bytes_t *in6) {
    ip_addr_t ip = { AF_INET6 };
    memset(&(ip.sa_data[0]),'\0',sizeof(ip.sa_data));
    *((struct in6_addr*) IP_ADDR_IN6(&ip)) = *((struct in6_addr*) in6);
    return ip;
}

in6_bytes_t ip_in6bytes_addr(register ip_addr_t *ip) {
    return *((in6_bytes_t*) IP_ADDR_IN6(ip));
}

ip_addr_t ip_addr_text(register const char *s) {
    ip_addr_t ip;
    memset(&ip,'\0',sizeof(ip));

    if (inet_pton(AF_INET6, s, ((struct in6_addr*) IP_ADDR_IN6(&ip))))
        ip.sa_family = AF_INET6;
    else if (inet_pton(AF_INET, s, ((struct in6_addr*) IP_ADDR_IN(&ip))))
        ip.sa_family = AF_INET;

    return ip;
}

static void hex_dump(uint8_t *src, uint32_t src_len, uint8_t *dst,
                     uint32_t dst_len, char sep) {
    while (src_len-- && (dst_len-=3))
        dst += sprintf((char*)dst, "%.2x%c", *(src++), sep);

    *(--dst)='\0';
}

char *ip_text_addr(register ip_addr_t *ip, register char *buf,
                   register size_t len) {
    static char sbuf[64] = "\0";
    char *d = buf ? buf : &(sbuf[0]);
    size_t sz = buf ? len : 64;
    char sep ='.';

    switch (ip->sa_family) {
        case AF_INET:
            return (char*) inet_ntop(AF_INET, (struct in_addr*) IP_ADDR_IN(ip), d, sz );
        case AF_INET6:
            return (char*) inet_ntop(AF_INET6, (struct in6_addr*) IP_ADDR_IN6(ip), d, sz );
#ifndef AF_LLC
#define AF_LLC 26
#endif
        case AF_LLC:
            sep = ':';
        default:
            hex_dump((uint8_t*) IP_ADDR_LLC(ip), sep == ':' ? 6 : __LLC_SOCK_SIZE__, (uint8_t*) d, sz, sep);
            return d;
    }

    return NULL;
}

char *ip_text(ip_addr_t ip, register char *buf, register size_t sz) {
    return ip_text_addr(&ip, buf, sz);
}

ip_addr_t ip_addr_binary(uint8_t *addr_data, uint8_t len) {
    ip_addr_t ip;
    uint8_t *addr = 0;
    memset(&ip, '\0', sizeof(ip_addr_t));

    switch (len) {
        case 4:
            ip.sa_family = AF_INET;
            addr = (uint8_t*) IP_ADDR_IN( &ip );
            break;
        case 6:
            ip.sa_family = AF_LLC;
            addr = (uint8_t*) IP_ADDR_IN6( &ip );
            break;
        case 16:
            ip.sa_family = AF_INET6;
            addr = (uint8_t*) IP_ADDR_LLC( &ip );
            break;
        default:
            return ip;
    }

    memcpy(addr, addr_data, len);
    return ip;
}

ip_addr_t ip_v4_broadcast(ip_addr_t *ipp, uint8_t prefix) {
    ip_addr_t ip = { AF_INET };
    memset(&(ip.sa_data[0]), '\0', sizeof(ip.sa_data));
    uint32_t ip4 = ntohl(((struct in_addr*) IP_ADDR_IN( ipp ))->s_addr);
    ((struct sockaddr_in*) &(ip))->sin_addr.s_addr= htonl((ip4 & (((1 << prefix) - 1) << (32 - prefix))) | (((1<< (32 - prefix)) - 1)));
    return ip;
}

uint8_t ip_v4_netmask_to_prefix(ip_addr_t *netmask) {
    if (netmask->sa_family != AF_INET)
        return 0;
    register uint32_t nm = ntohl(((struct sockaddr_in*) netmask)->sin_addr.s_addr), i = 32, b = 1;
    for(; i && ((nm & b) != b); i-=1, b <<= 1);
    return i;
}

ip_addr_t ip_v4_prefix_to_netmask(uint8_t prefix) {
    ip_addr_t ip = { .sa_family = AF_INET };
    ((struct sockaddr_in*) &ip)->sin_addr.s_addr = htonl(((1 << prefix) - 1) << (32 - prefix));
    return ip;
}

ip_addr_t ip_mask(ip_addr_t *ipA, uint8_t prefix) {
    ip_addr_t ip;
    ip = *ipA;

    switch (ipA->sa_family) {
        case AF_INET:
            {
                uint32_t a = ntohl(IP_ADDR_IN(ipA)->s_addr);
                prefix &= 0x1f;
                (IP_ADDR_IN(&ip))->s_addr = htonl(a & (((1 << prefix) - 1) << (32 - prefix)));
            }
            break;
        case AF_INET6:
            {
                uint8_t n = prefix & 0x7f,
                       *a = ((uint8_t*) IP_ADDR_IN6(&ip)),
                       *b = &(a[16]);

                for(; (b > a) && (n >= 8); b--, n-=8)
                    *b=0;

                if (n)
                    *b &= ((1 << n) - 1) << (8 - n);
            }
            break;
    }

    return ip;
}
